"""
a version of the peg game made playable with pygame.
"""
from __future__ import absolute_import
import os

import pygame

from .engine import Board

UPDATE_RATE=30

class PegHole(pygame.sprite.Sprite):
    
    SELECTED_COLOR=(255,0,0)
    OCCUPIED_COLOR=(0,0,0)
    EMPTY_COLOR=(255,255,255)
    TILE_HEIGHT=100
    TILE_WIDTH=100
    LEFT_OFFSET=60
    TOP_OFFSET=100
    CENTER_OFFSET=50
    RADIUS=15
    
    def __init__(self, board, point, selected=False):
        super(PegHole, self).__init__()
        self.board=board
        self.point=point
        self.selected=selected

        self.images=[]
        for c in (self.EMPTY_COLOR, self.OCCUPIED_COLOR, self.SELECTED_COLOR):
            s=pygame.Surface((50, 50))
            s.fill(PegGame.BACKGROUND_COLOR)
            pygame.draw.circle(s, c, s.get_rect().center, self.RADIUS)
            self.images.append(s.convert())
        self.update()
        self.rect=self.image.get_rect()
        x=self.point[1] * self.TILE_HEIGHT + self.TOP_OFFSET + (self.board.num_rows - self.point[0]) * self.CENTER_OFFSET
        y=self.point[0] * self.TILE_WIDTH + self.LEFT_OFFSET
        
        self.rect.topleft=(x,y)
        #print self.rect.topleft, self.point, 'occupied' if self.board.points[point] else 'empty'
        
    def update(self):
        if self.selected:
            self.image=self.images[2]
        elif self.board.points[self.point]:
            self.image=self.images[1]
        else:
            self.image=self.images[0]
    
    def has_position(self, pos):
        """
        returns whether or not the position is in the drawn circle.
        """
        cx, cy=self.rect.center
        x,y=pos
        # one grace pixel
        r=self.RADIUS + 1
        return (cx-r <= x <= cx+r) and (cy-r <= y <= cy+r)

class Panel(pygame.sprite.DirtySprite):
    """
    a blocking dialog that processes its own events.
    """
    def __init__(self, size):
        self.size=size
        self.image=pygame.Surface(size)
        self.draw()
        super(Panel, self).__init__()

    def draw(self):
        """
        put whatever you want on the background
        """
        pass

    def end_condition(self, evt):
        """
        redefine this to handle an event to dismiss the panel.
        By default, hitting escape, return, or Q will dismiss it.
        """
        return (evt.type==pygame.KEYDOWN and
                evt.key in (pygame.K_ESCAPE, pygame.K_q))

    def handle_event(self, evt):
        """
        override to provide any additional event handling
        """
        pass

    def run(self, screen, pos='center'):
        if pos == 'center':
            sr=screen.get_rect()
            ir=self.image.get_rect()
            pos=sr.centerx-ir.centerx, sr.centery-ir.centery
            del sr, ir
        original=screen.copy()
        r=screen.blit(self.image, pos)
        pygame.display.update()
        clock=pygame.time.Clock()
        running=True
        while running:
            clock.tick(UPDATE_RATE)
            for evt in pygame.event.get():
                if evt.type==pygame.QUIT:
                    pygame.quit()
                    return
                if self.end_condition(evt):
                    running=False
                    break
                else:
                    self.handle_event(evt)
        screen.blit(original, (0,0))
        pygame.display.update()


class HelpPanel(Panel):
    def __init__(self):
        super(HelpPanel, self).__init__((210,200))

    def end_condition(self, evt):
        """
        redefine this to handle an event to dismiss the panel.
        By default, hitting escape, return, or Q will dismiss it.
        """
        return (evt.type==pygame.KEYDOWN and
                evt.key in (pygame.K_ESCAPE, pygame.K_h, pygame.K_q))
    
    def draw(self):
        bg=(80,120,255)
        s=self.image
        s.fill(bg)

        help=['H: show/hide this help',
              'U: undo',
              'N: new game',
              'F: toggle fullscreen',
              'T: enable hints',
              'Q or Esc: quit']

        font = pygame.font.Font(None, 32)
        font.set_underline(True)
        text=font.render("Key bindings", 1, (255, 255, 255), bg)
        s.blit(text, (10, 10))
        font = pygame.font.Font(None, 24)
        for line, h in enumerate(help):
            text = font.render(h, 1, (255,255,255), bg)
            s.blit(text, (10, 42+(24 * line)))
        


class NonSound(object):
    def  play(self):
        pass
    
def load_sound(name):
    if pygame.mixer:
        return pygame.mixer.Sound(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                               'sounds',
                                               name))
    return NonSound()

        
class PegGame(object):
    BACKGROUND_COLOR=(190,190,190)

    def __init__(self, open='random'):
        self._open=open
        self.board=Board(open=open)
        self.holes=pygame.sprite.RenderPlain()
        for pt in self.board.points:
            self.holes.add(PegHole(self.board, pt))
        self.background=None
        self.sounds={}
        self.sounds['click']=load_sound('click.ogg')
        self.sounds['beep']=load_sound('beep.ogg')
        self.sounds['finish']=load_sound('finish.ogg')
        self.sounds['applause']=load_sound('applause.ogg')
        self.sounds['mockery']=load_sound('mockery.ogg')
        self.sounds['groan']=load_sound('groan.ogg')
        self._show_help=False
        self._showing_help=False
        self._help_screen=None
        self.show_moves=False
        self.movebox=pygame.Surface((100, 14 * 15))

        
    def draw_background(self, screen):
        self.background=pygame.Surface(screen.get_size()).convert()        
        self.background.fill(self.BACKGROUND_COLOR)
        font=pygame.font.SysFont('Verdana,Helvetica,SansSerif', 14)
        text=font.render('Press H for help.', 1, (0,170,0), self.BACKGROUND_COLOR)
        size=self.background.get_size()
        self.background.blit(text, (10, size[1]-24))
        screen.blit(self.background, (0,0))                             
        pygame.display.update()

    def draw_moves(self, screen):
        self.movebox.fill(self.BACKGROUND_COLOR)
        font = pygame.font.SysFont('Courier', 14)
        for line, move in enumerate(self.board.moves):
            text = font.render('%2d. %s %s' % (line + 1, move[0], move[1]), 1, (0,0,0), self.BACKGROUND_COLOR)
            self.movebox.blit(text, (10, 10+(14 * line)))
        screen.blit(self.movebox, (10, 10))
            

    def find_hole(self, pos):
        for h in self.holes:
            if h.has_position(pos):
                return h
        

    def update(self, screen):
        """
        this updates the display.
        """
        if self.background is None:
            self.draw_background(screen)
        self.draw_moves(screen)
        self.holes.update()
        rectlist=self.holes.draw(screen)
        pygame.display.update([x.rect for x in self.holes.sprites()] + [self.movebox.get_rect()])
        if self._show_help:
            if not self._showing_help:
                self.draw_help(screen)
                self._showing_help=True
        elif self._showing_help:
            self._help_screen.get_rect()
            

    def beep(self):
        self.sounds['beep'].play()
        #print 'BEEP'

    def click(self):
        # this sounds delayed, so commenting out ...
        #self.sounds['click'].play()
        #print "CLICK"
        pass

    def groan(self):
        self.sounds['groan'].play()


    def finish(self):
        if self.board.number==1:
            self.sounds['applause'].play()
        elif self.board.number <=4:
            self.sounds['finish'].play()
        else:
            self.sounds['mockery'].play()
        #print "DONE"

    def toggle_help(self):
        self._show_help=not self._show_help


def main():
    pygame.init()
    screen=pygame.display.set_mode((800, 600))
    pygame.display.set_caption("Smull-Peg")
    clock=pygame.time.Clock()

    # currently the parameters of this are fixed,
    # but it would be nice to be able to change the number of rows and
    # the open point.  @TBD
    game=PegGame()
    state='start'
    selected=None
    show_hints=False
    while True:
        clock.tick(UPDATE_RATE)
        for event in pygame.event.get():
            if (event.type==pygame.QUIT or
                (event.type==pygame.KEYDOWN and event.key in (pygame.K_ESCAPE, pygame.K_q))):
                pygame.quit()
                return
            elif event.type==pygame.KEYDOWN:
                
                if event.key==pygame.K_f:
                    pygame.display.toggle_fullscreen()
                elif event.key==pygame.K_h:
                    p=HelpPanel()
                    p.run(screen)
                    #game.toggle_help()
                elif event.key==pygame.K_u:
                    if game.board.moves:
                        game.board.undo()
                        state='start'
                        selected=None
                        for h in game.holes:
                            h.selected=False
                    else:
                        game.beep()
                elif event.key==pygame.K_t:
                     show_hints=not show_hints
                elif event.key==pygame.K_n:
                    state='start'
                    selected=None
                    game=PegGame()
            elif event.type==pygame.MOUSEBUTTONDOWN and event.button==1:
                pos=event.pos
                hole=game.find_hole(pos)
                if hole:
                    if state=='start':
                        # is this a legal move?
                        pt=hole.point
                        if pt in [x[0] for x in game.board.possible_moves]:
                            game.click()
                            hole.selected=True
                            selected=hole
                            state='selected'
                        else:
                            game.beep()
                    elif state=='selected':
                        if hole.selected:
                            game.click()
                            hole.selected=False
                            state='start'
                            selected=None
                        else:
                            pt=hole.point
                            assert selected
                            start_pt=selected.point
                            if not (start_pt, pt) in game.board.possible_moves:
                                game.beep()
                            else:
                                game.board.move(start_pt, pt)
                                game.click()

                                selected.selected=False
                                selected=None
                                if game.board.is_finished:
                                    game.finish()
                                    state='end'
                                else:
                                    if show_hints:
                                        winners=game.board.get_winning_moves()
                                        if winners:
                                            print "to win: %s" % ', '.join(["%s %s" % w for w in winners])
                                        else:
                                            game.groan()
                                            print "no winning moves"
                                    state='start'
                            
                        
        game.update(screen)


if __name__=='__main__':
    main()



"""
a model for the "I.Q. Tester" triangular peg board game.

"""
import anydbm
from copy import deepcopy
from operator import itemgetter
import os
import random
from string import center

try:
    import psyco
except ImportError:
    pass
else:
    psyco.full()

DB_PATH=os.path.join(os.path.abspath(os.path.dirname(__file__)), 'book.db')

class Point(tuple):
    """
    a named tuple Point class which can
    represents itself in a notation similar
    to algebraic notation in chess
    """

    def __new__(cls, *args):
        if len(args)==1:
            args=args[0]
        if isinstance(args, basestring):
            if len(args)!=2:
                raise ValueError("not a valid point: %s" % args)
            args=(ord(args[0])-97, int(args[1])-1)
        result=tuple.__new__(cls, args)
        if len(result)!=2:
            raise ValueError("not a valid point: %s" % result)
        return result


    def __str__(self):
        return '%s%d' % (chr(97+self[0]), 1+self[1])

    x=property(itemgetter(0))
    y=property(itemgetter(1))
    

class Board(object):
    """
    
    a representation of an I.Q. Tester triangular peg board game.  The
    holes of the board are represented by ordered pairs of the form
    (row, column) where both row and column start at zero at the top
    left.  The board stores the moves that have been played on it as
    well as the current position.

    Some important attributes and properties:

      start_open: the initial open point (by default randomly chosen).
      moves: the list of moves played.
      possible_moves: a list of legal moves from the given position.
      open_points: the points open on the board.
      closed_points: the points occupied by pegs.      
      number: the number of closed points.
      is_finished: if the game is finished or not (i.e., there are no possible moves).


    Example usage:

     >>> b=Board(open=(3,2))
     >>> b
         X          
        X X         
       X X X        
      X X O X       
     X X X X X     
     >>> b.move((1,0), (3,2))
         X          
        O X         
       X O X        
      X X X X       
     X X X X X     

    A board can be cloned (with clone()) if you want experiment with a
    position without disturbing your primary setup; the method clone_move()
    does that and also makes a move.

    To generate all solutions of a board that meet some criterion, use walk().

    """
    def __init__(self, num_rows=5, open='random'):
        self.num_rows=num_rows
        self.points=dict.fromkeys((Point(i,j) for i in range(num_rows) for j in range(i+1)), 1)
        if open=='random':
            open=random.choice(list(self.points))
        if open:
            self.start_open=open            
            self.open(open)
        self.moves=[]
        self._possible_moves=None

    def clone(self):
        """
        returns a copy of the board.
        """
        return deepcopy(self)

    def clone_move(self, p1, p2):
        """
        returns a copy of the board with the given move played.
        """
        return self.clone().move(p1, p2)

    def open(self, pt):
        pt=Point(pt)
        if not pt in self.points:
            raise ValueError("not a valid point for this board: %s" % pt)
        if self.start_open is None:
            self.start_open=pt
        self.points[pt]=0

    @property
    def open_points(self):
        return set(k for k, v in self.points.iteritems() if v==0)

    @property
    def closed_points(self):
        return set(k for k, v in self.points.iteritems() if v==1)

    @property
    def is_finished(self):
        return len(self.possible_moves)==0

    @property
    def number(self):
        return len(self.closed_points)

    def move(self, p1, p2):
        """
        move the peg at point p1 to point p2, removing the
        intervening peg.  Only legal moves allowed.  The move
        is recorded in the board's moves list, and the board itself
        is returned.
        """
        p3=self.removal_point(p1, p2)
        if not p3:
            raise ValueError("illegal move: %s => %s" % (p1, p2))
        self._possible_moves=None
        self.moves.append((Point(p1), Point(p2)))
        self.points[p1]=0
        self.points[p2]=1
        self.points[p3]=0
        return self

    def undo(self):
        if self.moves:
            u1,u2=m=self.moves.pop()
            self._possible_moves=None
            self.points[u1]=1
            self.points[u2]=0
            self.points[self.removal_point(u1, u2)]=1
            return m

    def ran_move(self):
        """
        make a random move.
        """
        if self.possible_moves:
            self.move(*random.choice(self.possible_moves))
        return self

    def ran_play(self):
        """
        continues playing from the current position, making
        random choices.
        """
        while True:
            moves=self.possible_moves
            if moves:
                m=random.choice(moves)
                self.move(*m)
            else:
                break
        return self

    def walk(self, save_criterion=lambda b: True):
        """
        an iterator returning all possible games from the current position
        that meet the save_criterion (defaulting to all of them).
        """
        if not self.possible_moves:
            if save_criterion(self):
                yield self
        else:
            for m in self.possible_moves:
                b=self.clone()
                b.move(*m)
                for r in b.walk(save_criterion):
                    yield r


    def __repr__(self):
        buff=['']
        width=self.num_rows*4
        for i in range(self.num_rows):
            line=[]
            for j in range(i+1):
                line.append('X' if self.points[(i,j)] else 'O')
            buff.append(center(' '.join(line), width))
        return '\n'.join(buff)

    @property
    def position_hash(self):
        return '%d%s' % (self.num_rows, ''.join(map(str, self.open_points)))


    def removal_point(self, p1, p2):
        """
        p2 must be open, p1 closed
        p1 and p2 must be on a line, separated by one closed point.
        returns the closed point.
        
        """
        if self.points[p2]:
            # p2 is occupied
            return None
        if not self.points[p1]:
            # p1 is empty
            return None
        if ((p1[0]==p2[0] and abs(p1[1]-p2[1])==2) or
            (p1[1]==p2[1] and abs(p1[0]-p2[0])==2) or
            (abs(p1[0]-p2[0])==2 and abs(p1[1]-p2[1])==2)):
            return Point(int((p1[0]+p2[0])/2), int((p1[1]+p2[1])/2))


    def lookup_winning_moves(self, path=DB_PATH):
        if not os.path.exists(path):
            return None
        db=anydbm.open(path)
        moves=db.get(self.position_hash, None)
        db.close()
        if moves is None:
            return None
        return [(Point(m[:2]), Point(m[2:])) for m in moves.split(',')] if moves else []    


    def get_winning_moves(self, use_book=True, path=DB_PATH):
        """
        returns the moves that lead to a solution with exactly one closed peg.
        N.B.: this can be expensive.
        """
        if use_book:
            ret = self.lookup_winning_moves(path=path)
            if ret is not None:
                return ret
        ret=[]
        for m in self.possible_moves:
            b=self.clone_move(*m)
            for r in b.walk(lambda x: x.number==1):
                ret.append(m)
                break
        return ret
    
    @property
    def possible_moves(self):
        if self._possible_moves is not None:
            return self._possible_moves
        moves=[]
        for p in self.open_points:
            # find pairs of closed points adjacent
            # to this point.  
            x, y=p
            for p1, p2 in (((x-1, y-1), (x-2, y-2)),  # NW
                           ((x-1, y),   (x-2, y)),    # NE
                           ((x,   y-1), (x,   y-2)),  # E
                           ((x,   y+1), (x,   y+2)),  # W
                           ((x+1, y),   (x+2, y)),    # SW
                           ((x+1, y+1), (x+2, y+2))): # SE

                if self.points.get(p1) and self.points.get(p2):
                    # both are closed, it is a move
                    moves.append((Point(p2), Point(p)))
        self._possible_moves=moves
        return moves
        
    
def play_random_game(num_rows=5, open=None):
    """
    plays a totally random game.
    """
    b=Board(num_rows=num_rows)
    if open is None:
        open=random.choice(list(b.points))
    b.open(open)
    print "open:", open
    print b
    print
    while True:
        possible=b.possible_moves
        if not possible:
            break
        m=random.choice(possible)
        b.move(*m)
        print "%s -> %s" % m
        print b
        print

    print "number of closed points: %d" % len(b.closed_points)
    
        



def _process_db(db, board, level, fail=False):
    print "in _process: level: %s, fail: %s" % (level, fail), board

    hash=board.position_hash
    if fail:
        moves=[]
        db[hash]=''
        db.sync()
    elif not hash in db:
        moves=board.get_winning_moves(False)
        db[hash]=','.join(['%s%s' % m for m in moves])
        db.sync()
    else:
        moves=db[hash]
        moves=[(Point(m[:2]), Point(m[2:])) for m in moves.split(',')] if moves else []
    
    if level > 0:
        for m in board.possible_moves:
            _process_db(db, board.clone_move(*m), level-1, not m in moves)

            

def build_db(levels=5, path=DB_PATH, num_rows=5):
    """
    this builds the book database.  Takes a long time.
    """
    db=anydbm.open(path, 'c')
    for p in list(Board(num_rows=num_rows).points):
        b=Board(open=p, num_rows=num_rows)
        _process_db(db, b, levels)
    db.close()
    



from distutils.core import setup

setup(name="smullpeg",
      version='0.1.0',
      description="a lamentable peg game",
      author="Jacob Smullyan",
      author_email="smulloni@smullyan.org",
      classifiers=['Topic :: Games/Entertainment :: Puzzle Games'],
      license="GPL",
      packages=["pegs"],
      package_data={'pegs' : ['book.db', 'sounds/*.ogg']},
      scripts=['smullpeg'])
